#ifndef RZB_ALERT_GLOBAL_H
#define RZB_ALERT_GLOBAL_H

#include <uuid/uuid.h>
#include "rzb_global.h"

/* Definitions to support ALERT_PACKET */
#define SHORT_DATA      0x00000001
#define LONG_DATA       0x00000002
#define FULL_BLOCK      0x00000004
#define NORM_BLOCK      0x00000008
#define SUB_COMPONENT   0x00000010
#define META_DATA       0x00000020
#define ALRT_MESSAGE    0x00000040     // NOT USED BY NORMAL ALERTING PROCESS
// Used to get the message string via
// the handleAlertBlob function

/* General Definitions */
#define INSERT_ALERT_LEN      200
#define MAX_MSG_SIZE          256
#define MAX_SHORT_SIZE        2048
#define MAX_LONG_SIZE         16777215


/* Protocol Additions to Support General Alerting */
typedef struct _ALERT_HEADER
{
    unsigned      alert_id;      // Filled in after db insert
    unsigned      timestamp;
    unsigned      event_id;      // Standard event data
    unsigned      nugget_id;
    unsigned short    priority;
    unsigned      src_ip;
    unsigned      dst_ip;
    unsigned short    ip_proto;
    unsigned short    src_port;
    unsigned short    dst_port;
    unsigned      flags1;        // Declare additional information available
    unsigned      flags2;        // reserved
    unsigned char     main_md5[MD5_SIZE];  // Subcompoments will have different_md5
    uuid_t            dataType;  // UUID of data
    unsigned char     db_md5[MD5_SIZE];
    unsigned      db_size;
    unsigned char     norm_md5[MD5_SIZE];
    unsigned      norm_size;
} ALERT_HEADER;


/* Protocol Additions to Support Component Management */
typedef struct _BLOB_REQUEST
{
    unsigned      alertID;
    unsigned      blob_type;
} BLOB_REQUEST;

typedef struct _BLOB_HEADER
{
    unsigned      alertID;
    unsigned      blobType;
    unsigned      blobSize;
    unsigned      flags; // Review for use
} BLOB_HEADER;

typedef struct _JUDGEMENT
{
    unsigned char md5[MD5_SIZE];
    unsigned size;
    unsigned flags;
} JUDGEMENT;

/* Structures to Support Alerting */
typedef struct _ALERT
{
    unsigned      event_id;
    unsigned short    priority;
    char src_ip[INET6_ADDRSTRLEN];
    char dst_ip[INET6_ADDRSTRLEN];
    unsigned short    ip_proto;
    unsigned short    src_port;            // NOT REQUIRED
    unsigned short    dst_port;            // NOT REQUIRED
    const char        *msg;
    unsigned      msg_size;
    const char        *short_data;
    unsigned      sd_size;
    uuid_t            dataType;
    unsigned char     *main_md5;           // NOT REQUIRED
    const char        *long_data;          // NOT REQUIRED
    unsigned      ld_size;             // NOT REQUIRED
    const void        *data_block;         // NOT REQUIRED
    unsigned      db_size;             // NOT REQUIRED
    const void        *norm_block;         // NOT REQUIRED
    unsigned      norm_size;           // NOT REQUIRED
} ALERT;

#endif /* RZB_ALERT_GLOBAL_H */

