#include <stdlib.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <dlfcn.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <uuid/uuid.h>
#include <openssl/evp.h>

#include "config.h"
#include "rzb_global.h"
#include "rzb_thread.h"
#include "rzb_api.h"
#include "rzb_network.h"
#include "rzb_cache.h"
#include "rzb_client.h"
#include "rzb_alert_util.h"
#include "rzb_client.h"
#include "rzb_thread.h"

static unsigned handlerCount = 0;
handlerNode *head = NULL;

static pthread_once_t once_cache = PTHREAD_ONCE_INIT;
static pthread_mutex_t cachemutex = PTHREAD_MUTEX_INITIALIZER;

static HRESULT addHandler(handler_fp fp)
{
    handlerNode *newNode;

    newNode = malloc(sizeof(*newNode));
    if (newNode == NULL)
        return R_MALLOC_FAIL;

    newNode->fp = fp;
    newNode->next = head;
    head = newNode;

    return R_SUCCESS;
}

static HRESULT updateNugget(const uuid_t nugtype, const uuid_t nugapp, const uuid_t datatype,
                            unsigned short nugport, unsigned nugid)
{
    //Connect to Dispatch Server
    RZBNetworkConnection *net = ConnectMe(nrtconfig.dsrvaddr, nrtconfig.dsrvport);
    if (!net)
    {
        printf("connection to %s:%s failed\n", nrtconfig.dsrvaddr, nrtconfig.dsrvport);
        return R_FAIL;
    }

    //Set up packet format to send
    SNAKECHARMER scpacket;
    NEWCAPPACKET newcap;
    protocolFill(&scpacket);
    uuid_copy(scpacket.datatype, nugtype);
    scpacket.msgtype = NEWCAP;

    /* Build Registration Packet */
    /* Nugget Network Info */
    newcap.nuggetid = nugid;
    inet_pton(AF_INET, nrtconfig.nugaddr, &newcap.nugaddr);
    newcap.nugport = nugport;

    /* Capability Declaration */
    uuid_copy(newcap.nugtype, nugtype);
    uuid_copy(newcap.nugapp, nugapp);
    uuid_copy(newcap.datatype, datatype);

    //Send packet
    if (sendWrap(net, &scpacket, sizeof(scpacket)) != R_SUCCESS)
    {
        NetworkConnectionDestroy(net);
        return R_FAIL;
    }

    if (sendWrap(net, &newcap, sizeof(newcap)) != R_SUCCESS)
    {
        NetworkConnectionDestroy(net);
        return R_FAIL;
    }

    //Recv ACK (hopefully)
    if (recvWrap(net, &scpacket, sizeof(scpacket)) != R_SUCCESS)
    {
        NetworkConnectionDestroy(net);
        return R_FAIL;
    }

    NetworkConnectionDestroy(net);
    return R_SUCCESS;
}

HRESULT rzbRegisterHandler(handler_fp fp, const uuid_t *acceptedTypes,
                           size_t numTypes, const uuid_t nuggetType)
{
    size_t i;
    static int hasReged = 0;
    static unsigned nuggetid = 0;

    handlerCount++;

    char uuid_str[SIZE_OF_UUID_STRING];

    // The stub will ignore types, but list what you attempted to register...
    printf("\nHandler %u accepts the following types:\n", handlerCount);
    for (i = 0; i < numTypes; i++)
    {
        if (hasReged == 0)
        {
            registerNugget(DETECTION,
                           nuggetType,
                           acceptedTypes[i],
                           strtoul(nrtconfig.nugport, NULL, 10),
                           nrtconfig.maximumthreads,
                           5,
                           nrtconfig.nugname,
                           &nuggetid);
            hasReged = 1;
        }
        else
            updateNugget(DETECTION, nuggetType, acceptedTypes[i], strtoul(nrtconfig.nugport, NULL, 10), nuggetid);
        uuid_unparse(acceptedTypes[i], uuid_str);
        printf("%s\n", uuid_str);
    }

    if (addHandler(fp) == R_FAIL)
        return R_FAIL;

    head->acceptedTypes = malloc(numTypes * sizeof(uuid_t));
    if (!head->acceptedTypes)
        return R_MALLOC_FAIL;

    memcpy(head->acceptedTypes, acceptedTypes, numTypes * sizeof(uuid_t));
    head->numTypes = numTypes;

    return R_SUCCESS;
}
/*
HRESULT     sendAlert(
                unsigned        alertid,
                char            *alertmsg,
                unsigned        metacode,
                unsigned        metasize,
                void            *metadata
            ) {

    printf("\tALERT - %d: %s\n", alertid, alertmsg);

    return (R_SUCCESS);
}
*/

static HRESULT checkResource(const uuid_t id, const char *data, const uuid_t type)
{
    printf("Requested a check on resource: %s\n", data);
    return(R_SUCCESS);
}

SO_PUBLIC HRESULT registerNugget(const uuid_t nugtype,
                                 const uuid_t nugapp,
                                 const uuid_t datatype,
                                 unsigned short nugport,
                                 unsigned instances,
                                 unsigned nugid,
                                 const char *name,
                                 unsigned *nuggetid)
{

    //Connect to Dispatch Server
    RZBNetworkConnection *net = ConnectMe(nrtconfig.dsrvaddr, nrtconfig.dsrvport);

    if (!net)
    {
        printf("connection to %s:%s failed\n", nrtconfig.dsrvaddr, nrtconfig.dsrvport);
        exit(-1);
    }

    //Set up packet format to send
    SNAKECHARMER scpacket;
    REGPACKET regpacket;

    /* Build Preamble Packet */
    protocolFill(&scpacket);
    scpacket.msgtype = REG;
    uuid_copy(scpacket.datatype, nugtype);

    /* Build Registration Packet */
    /* Nugget Network Info */
    inet_pton(AF_INET, nrtconfig.nugaddr, &regpacket.nugaddr);
    regpacket.nugport = nugport;
    regpacket.dummy = 0;

    /* Capability Declaration */
    uuid_copy(regpacket.nugtype, nugtype);
    uuid_copy(regpacket.nugapp, nugapp);
    uuid_copy(regpacket.datatype, datatype);
    regpacket.freeinstances = instances;
    regpacket.nuggetid = nugid;

    /* Nugget Name */
    memset(regpacket.name, 0, sizeof(regpacket.name));
    snprintf(regpacket.name, sizeof(regpacket.name), "%s", name);

    //Send packet
    if (sendWrap(net, &scpacket, sizeof(scpacket)) != R_SUCCESS)
    {
        NetworkConnectionDestroy(net);
        return R_FAIL;
    }

    if (sendWrap(net, &regpacket, sizeof(regpacket)) != R_SUCCESS)
    {
        NetworkConnectionDestroy(net);
        return R_FAIL;
    }

    //Recv ACK (hopefully)
    if (recvWrap(net, &scpacket, sizeof(scpacket)) != R_SUCCESS)
    {
        NetworkConnectionDestroy(net);
        return R_FAIL;
    }

    if (recvWrap(net, nuggetid, sizeof(*nuggetid)) != R_SUCCESS)
    {
        NetworkConnectionDestroy(net);
        printf("\nNo nugget id\n");
        return R_FAIL;
    }
    NetworkConnectionDestroy(net);
    //Check we got ACK from dispatch server
    //NEEDS WORK:
    //We need to grab a uuid generated by the dispatch server and return it
    //so we can use it in subsequent requests
    switch (scpacket.msgtype)
    {
        case ACK:
            return R_SUCCESS;

        default:
            return R_FAIL;
    }
}

/*
 *
 * Utility function that ships off data for detection
 *
 */
static void *sendDataStream(THREADARGS *threadargs)
{
    BLOCK_META_DATA *metaData = threadargs->metaData;

    unsigned id = metaData->eventid;
    unsigned char *data = metaData->data;
    unsigned datasize = metaData->size;
    uuid_t datatype;
    uuid_copy(datatype, metaData->datatype);
    unsigned moar = SENDBLOB;
    NetworkCacheResponse response = UNKNOWN_MD5_RESPONSE;

    unsigned result;

    pthread_once(&once_cache, initcache);

    ENTRY entry;
    int goodmd5result;
    int badmd5result;

    entry.size = datasize;
    md5sum(data, datasize, entry.d.chksum);

    pthread_mutex_lock(&cachemutex);
    goodmd5result = checkLocalEntry(&entry, GOODMD5);
    badmd5result = checkLocalEntry(&entry, BADMD5);
    pthread_mutex_unlock(&cachemutex);

    if (goodmd5result == R_FOUND)
    {
        printf("\n---CACHE ENTRY FOUND--- File is clean\n");
        if (metaData->data)
            free(metaData->data);
        free(metaData);
        unthreadme(threadargs);
        return NULL;
    }
    else if (badmd5result == R_FOUND)
    {
        printf("\n---CACHE ENTRY FOUND--- File is dirty\n");
        if (metaData->data)
            free(metaData->data);
        free(metaData);
        unthreadme(threadargs);
        return NULL;
    }

    result = remoteMD5CacheCheck(&id, entry.d.chksum, datasize, datatype, &response);

    if (result == R_SUCCESS)
    {
        if (response == KNOWN_GOOD_RESPONSE)
        {
            pthread_mutex_lock(&cachemutex);
            addLocalEntry(&entry, GOODMD5);
            pthread_mutex_unlock(&cachemutex);
            if (metaData->data)
                free(metaData->data);
            free(metaData);
            unthreadme(threadargs);
            return NULL;
        }
        else if (response == KNOWN_BAD_RESPONSE)
        {
            pthread_mutex_lock(&cachemutex);
            addLocalEntry(&entry, BADMD5);
            pthread_mutex_unlock(&cachemutex);
            if (metaData->data)
                free(metaData->data);
            free(metaData);
            unthreadme(threadargs);
            return NULL;
        }
    }
    else
        printf("\nFailed remote MD5 check\n");

    SNAKECHARMER *scpacket = calloc(1, sizeof(SNAKECHARMER) + sizeof(REQPACKET));

    //C is nuts, 1 means sizeof(SNAKECHARMER)
    REQPACKET *reqpacket = (REQPACKET *)(scpacket + 1);

    RZBNetworkConnection *net = ConnectMe(nrtconfig.dsrvaddr, nrtconfig.dsrvport);
    if (!net)
    {
        printf("connection to %s:%s failed\n", nrtconfig.dsrvaddr, nrtconfig.dsrvport);
        if (metaData->data)
            free(metaData->data);
        free(metaData);
        free(scpacket);
        unthreadme(threadargs);
        return NULL;
    }

    protocolFill(scpacket);
    scpacket->msgtype = REQ;
    uuid_copy(scpacket->datatype, datatype);

    reqpacket->filesize = datasize;
    reqpacket->eventid = id;
    metaData->eventid = id;

    if (sendWrap(net, scpacket, sizeof(SNAKECHARMER) + sizeof(REQPACKET)) != R_SUCCESS)
    {
        perror("Could not send file transfer request");
        free(scpacket);
        if (metaData->data)
            free(metaData->data);
        free(metaData);
        NetworkConnectionDestroy(net);
        unthreadme(threadargs);
        return NULL;
    }

    if (sendWrap(net, metaData, sizeof(*metaData)) != R_SUCCESS)
    {
        perror("Could not send file");
        free(scpacket);
        if (metaData->data)
            free(metaData->data);
        free(metaData);
        NetworkConnectionDestroy(net);
        unthreadme(threadargs);
        return NULL;
    }

    if (recvWrap(net, scpacket, sizeof(*scpacket)) != R_SUCCESS)
    {
        perror("Could not recv file transfer acknowledgement");
        free(scpacket);
        if (metaData->data)
            free(metaData->data);
        free(metaData);
        NetworkConnectionDestroy(net);
        unthreadme(threadargs);
        return NULL;
    }

    if (scpacket->msgtype != RST)
    {
        while (moar == SENDBLOB)
        {
            if (sendWrap(net, data, datasize) != R_SUCCESS)
            {
                perror("Could not send file");
                free(scpacket);
                if (metaData->data)
                    free(metaData->data);
                free(metaData);
                NetworkConnectionDestroy(net);
                unthreadme(threadargs);
                return NULL;
            }

            if (recvWrap(net, scpacket, sizeof(*scpacket)) != R_SUCCESS)
            {
                perror("Failed to receive additional instructions\n");
                free(scpacket);
                if (metaData->data)
                    free(metaData->data);
                free(metaData);
                NetworkConnectionDestroy(net);
                unthreadme(threadargs);
                return NULL;
            }
            moar = scpacket->msgtype;
        }
    }

    NetworkConnectionDestroy(net);
    free(scpacket);
    if (metaData->data)
        free(metaData->data);
    free(metaData);
    unthreadme(threadargs);

    return NULL;
}

SO_PUBLIC HRESULT sendData(BLOCK_META_DATA *metaData)
{
    THREADARGS *threadargs = malloc(sizeof(*threadargs));

    if (!(threadargs->metaData = metaData))
    {
        fprintf(stderr, "Failed to allocate thread arguments: %s\n", strerror(errno));
        return R_FAIL;
    }

    if (threadme(sendDataStream, threadargs) != R_SUCCESS)
    {
        if (metaData->data)
            free(metaData->data);
        free(metaData);
    }

    return R_SUCCESS;
}

// XXX OK, this should be calloc, but I don't have time to
// test before 0.1.  We will still have an init call as we
// use the /x00 fields to track whether or not fields are used
// XXX OK...then maybe we should calloc other fields where we
// have optional fields as null
BLOCK_META_DATA *initMetaData(void)
{
    return calloc(1, sizeof(BLOCK_META_DATA));
}

HRESULT AlertsDone(unsigned id)
{
    RZBNetworkConnection *net = ConnectMe(nrtconfig.dsrvaddr, nrtconfig.dsrvport);
    if (!net)
    {
        printf("connection to %s:%s failed\n", nrtconfig.dsrvaddr, nrtconfig.dsrvport);
        return R_FAIL;
    }

    SNAKECHARMER scpacket;

    protocolFill(&scpacket);
    scpacket.msgtype = DETECTDONE;

    if (sendWrap(net, &scpacket, sizeof(scpacket)) != R_SUCCESS)
    {
        perror("Could not send DETECTDONE packet");
        NetworkConnectionDestroy(net);
        return R_FAIL;
    }

    if (sendWrap(net, &id, sizeof(id)) != R_SUCCESS)
    {
        perror("Could not send DETECTDONE packet");
        NetworkConnectionDestroy(net);
        return R_FAIL;
    }

    NetworkConnectionDestroy(net);
    return R_SUCCESS;

}

static HRESULT sendMetaData(METAPACKET *mpacket, unsigned char *data)
{
    SNAKECHARMER scpacket;

    if (data == NULL || mpacket == NULL)
    {
        if (mpacket == NULL)
            perror("Uninitialized metapacket");
        else
            perror("Uninitialized data packet");
        return R_FAIL;
    }

    RZBNetworkConnection *net = ConnectMe(nrtconfig.dsrvaddr, nrtconfig.dsrvport);
    if (!net)
    {
        printf("connection to %s:%s failed\n", nrtconfig.dsrvaddr, nrtconfig.dsrvport);
        return R_FAIL;
    }

    mpacket->type = CAP_DETAILS;

    protocolFill(&scpacket);
    scpacket.msgtype = META;

    if (sendWrap(net, &scpacket, sizeof(scpacket)) != R_SUCCESS)
    {
        perror("Could not send scpacket");
        NetworkConnectionDestroy(net);
        return R_FAIL;
    }
    if (sendWrap(net, mpacket, sizeof(*mpacket)) != R_SUCCESS)
    {
        perror("Could not send metapacket");
        NetworkConnectionDestroy(net);
        return R_FAIL;
    }
    if (sendWrap(net, data, mpacket->size) != R_SUCCESS)
    {
        perror("Could not send metadata");
        NetworkConnectionDestroy(net);
        return R_FAIL;
    }
    if (recvWrap(net, &scpacket, sizeof(scpacket)) != R_SUCCESS)
    {
        perror("Could not receive scpacket");
        NetworkConnectionDestroy(net);
        return R_FAIL;
    }
    if (scpacket.msgtype != ACK)
        printf("Return message type: %u\n", scpacket.msgtype);

    NetworkConnectionDestroy(net);
    return R_SUCCESS;
}

static void handleSignal(int sig)
{
    stop_processing = 1;
}

SO_PUBLIC HRESULT rzbServer(const char *port, rzb_thread_func_t func, unsigned to_secs)
{
    int sock, conn;
    struct sockaddr_in dest_addr;
    socklen_t sockaddrsize;
    HRESULT retval;
    THREADARGS *threadargs;
    SNAKECHARMER scpacket;
    struct sigaction sa;
    RZBNetworkConnection *net;

    sock = MakeListener(port);

    printf("Now listening on port %s\n", port);

    memset(&sa, 0, sizeof(sa));
    sa.sa_handler = &handleSignal;
    sigaction(SIGINT, &sa, NULL);
    sigaction(SIGTERM, &sa, NULL);
    sigaction(SIGQUIT, &sa, NULL);

    retval = R_SUCCESS;
    while (!stop_processing)
    {
        sockaddrsize = sizeof(struct sockaddr_in);
        if ((conn = accept(sock, (struct sockaddr *)&dest_addr, &sockaddrsize)) == -1)
        {
            if (errno != EINTR)
            {
                retval = R_FAIL;
                fprintf(stderr, "Accept error: (%d) %s\n", errno, strerror(errno));
                perror("Accept error");
                break;
            }
        }
        else
        {
            net = MakeConnection(conn);
            if (net)
            {
                threadargs = calloc(1, sizeof(*threadargs));
                if (threadargs)
                {
                    threadargs->net = net;
                    retval = threadme(func, threadargs);
                }
                else
                    retval = R_FAIL;
            }
            else
            {
                threadargs = NULL;
                retval = R_FAIL;
            }
            if (retval == R_FAIL)
            {
                printf("\nThread dropped!\n");
                if (net)
                    sendFailure(net, &scpacket);
                NetworkConnectionDestroy(net);
                if (threadargs)
                    free(threadargs);
            }
        }
    }
    waitForIdle(to_secs);
    close(sock);
    return retval;
}

SO_PUBLIC const CollectionAPI rzb_collector =
{
    .checkResource = &checkResource,
    .sendData = &sendData,
    .sendMetaData = &sendMetaData,
    .registerNugget = &registerNugget,
};

