/************************************************
rzb_db.h -- Razorback Database Headerfile
Author -- Matthew Olney
************************************************/

#ifndef RZB_DB_H
#define RZB_DB_H

#include <uuid/uuid.h>
#include <mysql.h>
#include "rzb_global.h"
#include "rzb_nug_manage.h"

/* Define Dude */
#define UPDATE_CID_MID     "UPDATE components SET Metabook = '%u' WHERE Component_ID = '%u'"
#define GET_MID_CID        "SELECT Metabook FROM components WHERE Component_ID = %u"
#define CREATE_METABOOK    "INSERT INTO metabooks (Metabook_ID) VALUES (NULL)"
#define RETRIEVE_LAST_NUGID "select Nugget_ID from nuggets having Nugget_ID=(select max(Nugget_ID) from nuggets)"
#define RETRIEVE_DESCRIPTION_SQL "select Description from unique_ids where UID = '%s'"
#define FIND_PREREG_SQL "SELECT nugget_ID from nuggets where (IP = '%u' AND Port = '%d')"
#define INSERT_LOG_ENTRY   "INSERT INTO framework_logs (Log_Source, Log_Message, Log_Level) VALUES ('%s', '%s', %u)"
#define GET_NEW_EVENT_ID "INSERT INTO events (Event_ID, Timestamp) values (NULL, NOW())"

/* SQL ERROR CODES */
#define DUPLICATE_ENTRY 1062

/* SUPLLEMENTAL VALUES */
#define MAX_INT_STRING_SIZE 11

/* THREAT FLAGS */
#define KNOWN_GOOD   0x00000001
#define KNOWN_BAD    0x00000002
#define THREAT       0x00000004
#define HAZZARD      0x00000008
#define WHITE_LIST   0x00000010
#define TAINT        0x00000020
#define WATCH_LIST   0x00000040

/* Function Primatives */
void printError(const char *errMessage, MYSQL *conn);
HRESULT getUUIDDesc(const char *UUID, char *desc);
HRESULT printUUIDList(uuid_t *list, unsigned listSize);
HRESULT buildUUIDList(unsigned *count, uuid_t **newList, const char *request);
HRESULT dbRegisterNugget(NUGGETENTRY *newNug);
HRESULT insertLog(unsigned level, const char *msg, const char *src);
HRESULT protectString(const char *data, unsigned size, char **protected_data);
HRESULT protectHex(const unsigned char *data, unsigned size, char **protected_data);
HRESULT getEventID(unsigned *event_id);
HRESULT getMIDByCID(unsigned component_id, unsigned *metabook_id);
HRESULT insertComponentMetaData(METAPACKET *mpacket, const unsigned char *data, unsigned size);
HRESULT insertMIDByCID(unsigned component_id, unsigned metabook_id);
HRESULT createMetaBook(unsigned *metabook);
HRESULT checkMD5(const unsigned char *md5, unsigned size, int *cache_result);

MYSQL *getMYSQLConnection(void);
void putMYSQLConnection(MYSQL *conn);
void destroyMYSQLConnections(void);

#endif /* RZB_DB_H */

