#include <mysql.h>
#include <syslog.h>
#include <uuid/uuid.h>

#include "config.h"
#include "rzb_global.h"
#include "rzb_db.h"
#include "rzb_component_db.h"
#include "rzb_workstation_db.h"
#include "rzb_workstation_global.h"
#include "rzb_network.h"
#include "rzb_management_log.h"

static const char GET_LOGS_BY_NUM[] = "SELECT Log_ID, Log_Source, Log_Message, unix_timestamp(Timestamp), Log_Level FROM framework_logs ORDER BY Log_ID DESC LIMIT %u";
//static const char GET_LAST_LOG_ID[] = "SELECT Log_ID FROM framework_logs ORDER BY Log_ID DESC LIMIT 1";

HRESULT fetchLogsByNumber(RZBNetworkConnection *net, unsigned num)
{
    MYSQL *conn;
    LOG_RESPONSE_PACKET pkt;
    MYSQL_RES *logs;
    MYSQL_ROW row;
    char *request;
    size_t request_size;

    request_size = sizeof(GET_LOGS_BY_NUM) + MAX_INT_STRING_SIZE;
    if ((request = malloc(request_size)) == NULL)
    {
        submitLog(LOG_ERR, "Failed to allocate memory for request", "fetchLogsByNumber");
        return R_FAIL;
    }
    snprintf(request, request_size, GET_LOGS_BY_NUM, num);

    if ((conn = getMYSQLConnection()) == NULL)
    {
        free(request);
        return R_FAIL;
    }
    if (mysql_query(conn, request) != 0)
    {
        submitLog(LOG_ERR, "Failed to get query response", "fetchLogsByNumber");
        free(request);
        putMYSQLConnection(conn);
        return R_DB_FAIL;
    }
    free(request);

    if (!(logs = mysql_store_result(conn)))
    {
        submitLog(LOG_ERR, "Failed to get logs", "fetchLogsByNumber");
        putMYSQLConnection(conn);
        return R_FAIL;
    }

    while ((row = mysql_fetch_row(logs)) != NULL)
    {
        pkt.id = strtoul(row[0],NULL,10);
        pkt.timestamp = (unsigned)strtoul(row[3], NULL, 10);
        pkt.level = (unsigned)strtoul(row[4], NULL, 10);
        pkt.src_size = (unsigned)strlen(row[1]);
        pkt.msg_size = (unsigned)strlen(row[2]);

        if (sendWrap(net, &pkt, sizeof(pkt)) != R_SUCCESS)
        {
            submitLog(LOG_ERR, "Failed to send log response", "fetchLogsByNumber");
            mysql_free_result(logs);
            putMYSQLConnection(conn);
            return R_FAIL;
        }
        if (sendWrap(net, row[1], strlen(row[1])) != R_SUCCESS)
        {
            submitLog(LOG_ERR, "Failed to send log response", "fetchLogsByNumber");
            mysql_free_result(logs);
            putMYSQLConnection(conn);
            return R_FAIL;
        }
        if (sendWrap(net, row[2], strlen(row[2])) != R_SUCCESS)
        {
            submitLog(LOG_ERR, "Failed to send log response", "fetchLogsByNumber");
            mysql_free_result(logs);
            putMYSQLConnection(conn);
            return R_FAIL;
        }
    }
    pkt.id = 0;
    sendWrap(net, &pkt, sizeof(pkt));
    mysql_free_result(logs);
    putMYSQLConnection(conn);
    return R_SUCCESS;
}

