#include <uuid/uuid.h>
#include <unistd.h>
#include <getopt.h>
#include <time.h>
#include <arpa/inet.h>

#include "config.h"
#include "rzb_global.h"
#include "rzb_api.h"
#include "rzb_conf.h"
#include "rzb_client.h"
#include "rzb_network.h"
#include "rzb_thread.h"

#define DSTIP "192.168.2.1"
#define SRCIP "192.168.3.1"

static void usage(const char *name)
{
    printf("Usage: %s --type=<file_type> --file=<file_name> [--conf=<configuration file>]\n", name);
    printf("File types:\n");
    printf("PDF_FILE\n");
    printf("SWF_FILE\n");
    printf("ZIP_FILE\n");
    printf("PE_FILE\n");
    printf("OLE_DOC\n");
    printf("MAIL_CAPTURE\n");
    printf("WEB_CAPTURE\n");
    printf("DNS_CAPTURE\n");
    printf("SHELLCODE\n");

    exit(-1);
}

int main(int argc, char ** argv)
{
    BLOCK_META_DATA *metaData;
    uint8_t *file;
    //unsigned nuggetid = 0;
    char src_ip_str[INET6_ADDRSTRLEN];
    char dst_ip_str[INET6_ADDRSTRLEN];
    char uuid_str[SIZE_OF_UUID_STRING];
    //unsigned i;
    struct stat sb;
    uuid_t nugType;
    FILE *f;

    static const struct option long_options[] =
    {
        {"conf",required_argument, NULL, 'c'},
        {"type",required_argument, NULL, 't'},
        {"file",required_argument, NULL, 'f'},
        {NULL, 0, NULL, 0}
    };

    int optId; /* holds the option index for getopt*/
    const char *configFile = "rzb.conf";
    const char *fileType = NULL;
    const char *fileName = NULL;
    int option_index = 0;

    while (1)
    {
        optId = getopt_long(argc, argv, "c:t:f:", long_options, &option_index);

        if (optId == -1)
            break;

        switch (optId)
        {
            case 'c':
                configFile = optarg;
                break;
            case 't':
                fileType = optarg;
                break;
            case 'f':
                fileName = optarg;
                break;
            case '?':  /* getopt errors */
                break;
            default:
                break;
        }
    }

    srand(time(NULL));

    if (!fileType || !fileName)
        usage(argv[0]);

    readConfig(&nrtconfig, configFile);

    if (!strcmp(fileType, "PDF_FILE"))
        uuid_copy(nugType, PDF_FILE);
    else if (!strcmp(fileType, "SWF_FILE"))
        uuid_copy(nugType, SWF_FILE);
    else if (!strcmp(fileType, "PE_FILE"))
        uuid_copy(nugType, PE_FILE);
    else if (!strcmp(fileType, "OLE_DOC"))
        uuid_copy(nugType, OLE_DOC);
    else if (!strcmp(fileType, "MAIL_CAPTURE"))
        uuid_copy(nugType, MAIL_CAPTURE);
    else if (!strcmp(fileType, "WEB_CAPTURE"))
        uuid_copy(nugType, WEB_CAPTURE);
    else if (!strcmp(fileType, "DNS_CAPTURE"))
        uuid_copy(nugType, DNS_CAPTURE);
    else if (!strcmp(fileType, "ZIP_FILE"))
        uuid_copy(nugType, ZIP_FILE);
    else if (!strcmp(fileType, "SHELLCODE"))
        uuid_copy(nugType, SHELLCODE);
    else
        usage(argv[0]);

    if (uuid_is_null(nugType))
    {
        perror("Error converting uuid type");
        exit(-1);
    }

    uuid_unparse(nugType, uuid_str);
    //printf("Registering for type %s\n", uuid_str);
    //registerNugget(COLLECTOR, TESTCOLLECT, nugType, (uint16_t)((rand() % 30000) + 30000), 1, 0, "dc-nf-02", &nuggetid);

    if ((f = fopen(fileName, "rb")) == NULL)
    {
        fprintf(stderr, "Error opening input file: %s\n", fileName);
        exit(-1);
    }

    if (fstat(fileno(f), &sb) == -1)
    {
        perror("fstat");
        exit(-1);
    }

    printf("Allocating file buffer\n");
    if ((file = malloc(sb.st_size)) == NULL)
    {
        fprintf(stderr, "Failed to allocate %ld bytes: %s\n", sb.st_size, strerror(errno));
        exit(-1);
    }
    printf("Reading file\n");
    if (fread(file, sb.st_size, 1, f) != 1)
    {
        fprintf(stderr, "Error reading input file %s: %s\n", fileName, strerror(errno));
        exit(-1);
    }
    fclose(f);

    if ((metaData = calloc(1, sizeof(BLOCK_META_DATA))) == NULL)
    {
        perror("Error allocating BLOCK_META_DATA");
        exit(-1);
    }

    // Populate metadata required fields
    metaData->timestamp = time(NULL);
    metaData->data = file;
    metaData->size = sb.st_size;
    uuid_copy(metaData->datatype, nugType);
    inet_pton(AF_INET, DSTIP, &metaData->dst_ip);
    inet_pton(AF_INET, SRCIP, &metaData->src_ip);
    inet_ntop(AF_INET, &metaData->src_ip, src_ip_str, sizeof(src_ip_str));
    inet_ntop(AF_INET, &metaData->dst_ip, dst_ip_str, sizeof(dst_ip_str));
    printf("Src: %s  Dst: %s\n", src_ip_str, dst_ip_str);

    // Optional metadata fields
    metaData->ip_proto = 6;
    metaData->src_port = 24391;
    metaData->dst_port = 80;

    printf("Calling senddata\n");
//    for(i=0;i<100;i++){
    sendData(metaData);
//  }
    waitForIdle(30);

    return 0;
}
